% =========================================================================
% Title       : Complex Reduction of Amplitudes (CRAM)
% File        : CRAM.m
% -------------------------------------------------------------------------
%
% Description:
%
%   This method solves the following optimization problem 
%
%       minimize:   ||u||_inf
%     subject to:   ||Du-b||_2 < eps
%
%   by reformulating it as the following saddle-point problem:
%
%     max_y min_x ||x1||_inf + <y,Dx1-b-x2> subject to ||x2||_2<eps
% 
%   which is solved by pdhg_adaptive.m. CRAM was described in: 
%
%      C. Studer, T. Goldstein, W. Yin, and R. G. Baraniuk, "Democratic 
%      Representations," arXiv preprint: 1401.3420, 2014
%
% Usage:
%
%   [ u, outs ] = CRAM(D,Dt,b,eps)
%
% Inputs:
%
%     D and Dt  : Function handles to an MxN matrix D and its adjoint                      
%     eps       : stopping tolerance
%     max_iter  : maximum number of iterations
%
% Outputs:
%
%     u        : the solution
%     outs     : a struct containing various outputs and convergence data
%
% ------------------------------------------------------------------------- 
%
% Revisions:
%
%   Date       Version  Author     Description
%   Jun-08-15  1.0      goldstein  code release
%
% -------------------------------------------------------------------------
% (C) 2015 Tom Goldstein (e-mail: tomg@cs.umd.edu)
% =========================================================================

function [ u, outs ] = CRAM(D,Dt,b,eps,max_iter)

%% Check preconditions and setup the problem

%  Test that b is a column vector
assert(numel(b) == size(b,1), 'Input b must be a column vector');

%  If D and Dt are matrices, make function handles
if isnumeric(D)
    Dt = @(x) D'*x;
    D = @(x) D*x;
end

% Check that Dt works right.  Get approximate reconstruction
uApprox = Dt(b);
assert(numel(uApprox) == size(uApprox,1), 'Dt must return a column vector');

%  Scale solution to get faster convergence
scale = norm(uApprox)/norm(b)^2;
b = b*scale;
eps = eps*scale;
uApprox = uApprox*scale;

%  Get the problem dimensions
M = size(b,1);
N = size(uApprox,1);

%  Test that operators are Hermitian adjoints
u = randn(N,1);
v = randn(M,1);
assert(abs(D(u)'*v-u'*Dt(v))<1e-9, 'D and Dt are not Hermitian adjoint');

%% Define the ingredients we need for PDHG

%  This matrix computes Dx1-x2
A = @(x) D(x(1:N)) - x(N+1:end);
%  The transpose of A
At = @(y) [Dt(y); - y];

% This operator does both infinity norm prox on x1, and projection of
% x2 onto the l2 ball of radius epsilon
xProx = @(x, tau) [prox_inf(x(1:N),tau); projectIntoCircle(x(N+1:end),eps)];
% Handles the extra term involving b in the energy
yProx = @(y,sigma) y-sigma*b;

% Initial iterates
x0 = zeros(N+M,1); % Contains both x1 and x2
y0 = zeros(M,1);   % Lagrange multipliers to enforce Dx1-b-x2=0

%% Call the solver
% The options that go into adaptive PDHG
opts = [];
opts.eps = norm(uApprox)/N*1e-2; % Stopping tolerance proportional to ||b||
opts.maxIters = max_iter;

%  Call the solver
[x,outs]= pdhg_adaptive(x0,y0,A,At,xProx,yProx,opts);

%%  Get the result, and divide out the scaling
u = x(1:N)/scale;

end

%%  Projext the vector x into a circle with radius 'rad'
function x = projectIntoCircle(x, rad)

n = norm(x,'fro');
if n>rad
    x = x/n*rad;
end

end
