% =========================================================================
% Purpose     : Test various l_inf-norm solvers 
% File        : test_solvers.m
% -------------------------------------------------------------------------
%
% Description:
%
%   This method solves a basic example problem using FITRA, 
%   CRAM, and CRAMP, respectively:
%
%       minimize:   lambda*||u||_inf + 0.5*||Du-b||_2
%
%       minimize:   ||u||_inf
%     subject to:   ||Du-b||_2 < eps
%
%       minimize:   ||u||_inf
%     subject to:   Du=b
% 
%   FITRA was described in 
%
%      C. Studer and E. G. Larsson, "PAR-aware large-scale multi-user 
%      MIMO-OFDM downlink." IEEE Journal on Selected Areas in 
%      Communications, 2013
%      
%   CRAM and CRAMP were described in 
%
%      C. Studer, T. Goldstein, W. Yin, and R. G. Baraniuk, "Democratic 
%      Representations," arXiv preprint:1401.3420, 2014
%
% Usage:
%
%   Just run the script and enjoy!
%
% ------------------------------------------------------------------------- 
%
% Revisions:
%
%   Date       Version  Author     Description
%   Jun-08-15  1.0      studer     code release
%
% -------------------------------------------------------------------------
% (C) 2015 Tom Goldstein (e-mail: tomg@cs.umd.edu)
% (C) 2015 Christoph Studer (e-mail: studer@cornell.edu)
% =========================================================================

clc
rng(0); % Seed the random number generator to make results repeatable

% -- setup the problem

fprintf('\nSetting up problem...\n\n');
M = 500;  % length of vector being represented
N = 2000; % length of representation
max_iter = 500;  % Maximum iterations of solvers

% The signal to be represented - use a random vector
b = randn(M,1);

% Create a random DCT ensemble (i.e., a Parseval frame)
% The target signal will be represented using columns from this frame
D = dct(eye(N));
idx = randperm(N);
D = D(idx(1:M),:);
Dt = @(x) D'*x;
D = @(x) D*x;
 
% -- SUPER HAPPY FUN OPTIMIZATION TIME GO!!

% call FITRA
fprintf('Calling the FITRA solver...');
lambda = 3; 
tic
[u1] = FITRA(D,Dt,b,N,lambda,max_iter);
fprintf('Done. ');
toc

% call CRAM
fprintf('Calling the CRAM  solver...');
eps = norm(b)*1e-6; % The constraint is ||Dx-b||<eps
tic
[u2] = CRAM(D,Dt,b,eps,max_iter);
fprintf('Done. ');
toc

% call CRAMP
fprintf('Calling the CRAMP solver...');
tic
[u3] = CRAMP(D,Dt,b,max_iter);
fprintf('Done. ');
toc

% -- plot the results
figure(1)
subplot(3,1,1), plot(u1); title('FITRA')
subplot(3,1,2), plot(u2); title('CRAM')
subplot(3,1,3), plot(u3); title('CRAMP')


% -- display the approximation error
fprintf('\nFITRA Approximation Error = %f\n',norm(D(u1)-b)/norm(b));
fprintf('CRAM  Approximation Error = %f\n',norm(D(u2)-b)/norm(b));
fprintf('CRAMP Approximation Error = %f\n\n',norm(D(u3)-b)/norm(b));

% -- display the PAPR
fprintf('FITRA PAPR = %f\n',N*norm(u1,inf)^2/norm(u1,2)^2);
fprintf('CRAM  PAPR = %f\n',N*norm(u2,inf)^2/norm(u2,2)^2);
fprintf('CRAMP PAPR = %f\n',N*norm(u3,inf)^2/norm(u3,2)^2);
